; fzs.asm - Serielle Schnittstelle für Frequenzzähler FZ1310
;
; Copyright (C) 2012 - 2015 Nils Kassube <kassube@gmx.net>
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License version 2 as
; published by the Free Software Foundation.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
; USA.

.equ	fosc = 18432000		; Quarzfrequenz in Hz
.equ	BAUD = 115200		; Baudrate für serielle Schnittstelle

	.nolist
.include "tn2313def.inc"
	.list

.equ	STROBE		= $80
.equ	TAB		= $09
.equ	CR		= $0d
.equ	LF		= $0a
.equ	MINPROTAG	= 24*60
.equ	MSPROMIN	= 60*1000

.def	zero		= r0
.def	zeroh		= r1
.def	temp		= r2
.def	nl		= r4
.def	nh		= r5
.def	dl		= r6
.def	dh		= r7
.def	ql		= r8
.def	qh		= r9
.def	rl		= r10
.def	rh		= r11
.def	n		= r12
.def	d		= r13
.def	q		= r14
.def	r		= r15
.def	w1		= r16
.def	w2		= r17
.def	w3		= r18
.def	w4		= r19
.def	ff		= r20
.def	cnt		= r21
.def	up0		= r22
.def	up1		= r23
.def	up2		= r24
.def	up3		= r25

	.dseg
	.org	SRAM_START
sign:	.byte	1
a1:	.byte	1
a2:	.byte	1
a3:	.byte	1
a4:	.byte	1
a5:	.byte	1
a6:	.byte	1
a7:	.byte	1
a8:	.byte	1
expmsb:	.byte	1
explsb:	.byte	1

	.cseg
	.org	$00
	rjmp	RESET
	.org	$0d
	rjmp	TIMER0_COMPA		; Timer0 Compare A Interrupt Handler

	.org	$13
TIMER0_COMPA:
	in	temp, sreg
	inc	up0
	brne	t1
	inc	up1
t1:
	cpi	up0, low(MSPROMIN)
	brne	t2
	cpi	up1, high(MSPROMIN)
	brne	t2
	movw	up1:up0, zeroh:zero
	inc	up2
	brne	t2
	inc	up3
t2:
	out	sreg, temp
	reti

RESET:
	ldi	w1, low(RAMEND)
	out	SPL, w1
	clr	zero
	clr	zeroh
	ser	ff
;
; Uhr (Uptime) initialisieren
;
	movw	up1:up0, zeroh:zero
	movw	up3:up2, zeroh:zero

; Mit Timer0 Takt generieren für Uhr-Interrupt mit 1kHz
; CTC-Mode -> WGM01 = 1
	ldi	w1, (1<<WGM01)			; CTC Mode
	out	TCCR0A, w1
;
;	CS02	CS01	CS00	Description
;	0	0	0	No clock source / Timer stopped
;	0	0	1	N0 = 1
;	0	1	0	N0 = 8
;	0	1	1	N0 = 64
;	1	0	0	N0 = 256
;	1	0	1	N0 = 1024
;	1	1	0	External clock, falling edge (pin T0)
;	1	1	1	External clock, rising edge (pin T0)
;
	ldi	w1, (1<<CS02) | (1<<CS00)	; N0 = 1024
	out	TCCR0B, w1

; fint1 = fosc / (N0 * (1 + OCR0A))
; OCR1A = fosc / (N0 * fint0) - 1

.equ	N0 = 1024
.equ	fint0 = 1000
.equ	xocr0a = fosc / (N0 * fint0) -1
.if	xocr0a > $ff
.error "xocr0a zu gross!"
.endif
	ldi	w1, xocr0a
	out	OCR0A, w1
	ldi	w1, (1<<OCIE0A)
	out	TIMSK, w1
	sei
;
; Serielle Schnittstelle initialisieren
;
.equ	xubrr = fosc / (16 * BAUD) -1
	ldi	w1, low(xubrr)
.if	xubrr > $ff
	ldi	w2, high(xubrr)
	out	UBRRH, w2
.else
	out	UBRRH, zero
.endif
	out	UBRRL, w1
;
; Default frame format: 8data, 1stop bit
;	ldi	w1, (3<<UCSZ0)
;	out	UCSRC,w1

	ldi	w1, (1<<RXEN) | (1<<TXEN)
	out	UCSRB, w1
;
; Grundeinstellung Ports
;
	sbi	PORTD, PD4			; /OE -> 1
	sbi	DDRD, PD4
	cbi	PORTD, PD6
	sbi     DDRD, DDD6
	out	PORTB, ff
	out	DDRB, ff
	sbi	PORTD, PD6
	cbi	PORTD, PD6
	out	DDRB, zero
	out	PORTB, zero

loop:
	sbis	UCSRA, RXC
	rjmp	loop
	in	w1, UDR
;
; Befehle filtern
;
	cpi	w1, $31
	brge	loop1
	cpi	w1, $05
	brlo	loop0
	cpi	w1, $10
	brlo	err
	cpi	w1, $21
	breq	err
	cpi	w1, $2c
	breq	err
	cpi	w1, $2e
	brlo	loop0
	cpi	w1, $30
	brne	err
;
; Befehl an FZ1310 weitergeben
; und Ergebnis zurücksenden
;
loop0:
	rcall	io
	mov	w1, w2
	rcall	tx
	rjmp	loop
;
; Sonderbehandlung zusätzliche Befehle
; sonst Fehlermeldung
;
loop1:
	cpi	w1, '@'
	breq	dm
	cpi	w1, '?'
	breq	menu
err:
	ldi	ZL, low(2*help1)
	ldi	ZH, high(2*help1)
err1:
	rcall	print
	rjmp	loop

menu:
	ldi	ZL, low(2*help2)
	ldi	ZH, high(2*help2)
	rjmp	err1

;
; Dauermessung
;
dm:
	ldi	w1, $30
	rcall	io
dm1:
	ldi	w1, $20			; auf Messergebnis warten
	rcall	io
	tst	w2			; 0 = Messung beendet
	brne	dm1
	cli
	movw	XH:XL, up1:up0		; Uhrzeit merken
	movw	YH:YL, up3:up2
	sei
	clr	ZH
	ldi	ZL, low(a1)
	ldi	w1, $22
dm2:
	rcall	io
	st	Z+, w2
	inc	w1
	cpi	ZL, low(explsb+1)	; LSB Exponent = Ende
	brlo	dm2
	ldi	w1, $2d			; Vorzeichen
	rcall	io
	sts	sign, w2
	ldi	w1, $30			; neue Messung starten
	rcall	io
;
; Ausgabe der Messung
;
	lds	w1, sign
	andi	w1, $01
	ldi	w1, ' '
	breq	dm3
	ldi	w1, '-'
dm3:
	rcall	tx
;	clr	ZH
	ldi	ZL, low(a1)
dm4:
	ld	w1, Z+
	rcall	prnum
	cpi	ZL, low(a2)
	brne	dm5
	ldi	w1, '.'
	rcall	tx
dm5:
	cpi	ZL, low(expmsb)
	brlo	dm4
	ldi	w1, 'e'
	rcall	tx
	lds	w1, expmsb
	andi	w1, $0f
	swap	w1
	lds	w2, explsb
	andi	w2, $0f
	or	w2, w1
	tst	w2
	ldi	w1, '+'
	brpl	dm6
	ldi	w1, '-'
dm6:
	rcall	tx
	mov	w1, w2
	tst	w1
	brpl	dm7
	clr	w1
	sub	w1, w2
dm7:
	rcall	pr2num
;
; Uhrzeit (Uptime) ausgeben
;
; YH:YL = 0d 00h:00m - 45d 12h:15m
; XH:XL = 00.000 s - 59.999 s
;
	ldi	w1, TAB
	rcall	tx
	movw	dh:dl, YH:YL
	ldi	w1, low(MINPROTAG)
	ldi	w2, high(MINPROTAG)
	movw	nh:nl, w2:w1
;
; YH:YL / MINPROTAG
; -> Tage in qh:ql (da max. 45 reicht ql)
;    Stunden/Minuten in rh;rl
;
	rcall	div16
	mov	w1, ql
	rcall	pr2num
	ldi	w1, 'd'
	rcall	tx
	ldi	w1, ' '
	rcall	tx
;
; Stunden/Minuten teilen
; -> Stunden in qh;ql (da max. 23 reicht ql)
;    Minuten in rh:rl (da max. 59 reicht rl)
;
	movw	dh:dl, rh:rl
	clr	nh
	ldi	w1, 60
	mov	nl, w1
	rcall	div16
	mov	w1, ql
	rcall	pr2num
	ldi	w1, ':'
	rcall	tx
	mov	w1, rl
	rcall	pr2num
	ldi	w1, ':'
	rcall	tx
;
; Aus XH:XL ganze Sekunden berechnen
; -> Sekunden in qh:ql (da max. 59 reicht ql)
;    Millisekunden in rh:rl
;
	movw	dh:dl, XH:XL
	ldi	w1, low(1000)
	ldi	w2, high(1000)
	movw	nh:nl, W2:w1
	rcall	div16
	mov	w1, ql
	rcall	pr2num
	ldi	w1, '.'
	rcall	tx
;
; Millisekunden ausgeben
;
	movw	dh:dl, rh:rl
	clr	nh
	ldi	w1, 10
	mov	nl, w1
	rcall	div16
	mov	w1, ql
	rcall	pr2num
	mov	w1, rl
	rcall	prnum
	rcall	newline
	sbis	UCSRA, RXC	; (beliebiges) Zeichen eingegeben?
	rjmp	dm1
	in	w1, UDR		; Zeichen abholen (nicht auswerten)
	rjmp	loop		; Dauermessung beendet

iotimeout:
	rcall	stop500ms
;
; Lege Befehl an IN 0-6
; Setze /READY auf LOW
	clc
	rcall	putcmd
; warte 500ms
	rcall	delay500ms
; Setze /READY auf HIGH
	sec
	rcall	putcmd
; warte 500ms
	rcall	delay500ms
; Springe nach START-IN-OUT-UNTERPROGRAMM
;	rjmp	io
;
io:
; START-IN-OUT-UNTERPROGRAMM
; Befehl in w1
; Ergebnis in w2
;
; Setze /READY auf HIGH
;
	sec
	rcall	putcmd
;
; Warte auf /STROBE = HIGH, springe nach 500ms nach TIMEOUT
;
	sec
	rcall	strb
	brcs	iotimeout
;
; Lade Befehl an IN0-6
; Setze /READY auf LOW
;
	clc
	rcall	putcmd
;
; Warte auf /STROBE = LOW, springe nach 500ms nach TIMEOUT
;
	clc
	rcall	strb
	brcs	iotimeout
;
; Setze /READY auf HIGH
;
	sec
	rcall	putcmd
; 
; Warte auf /STROBE = HIGH, springe nach 500ms nach TIMEOUT
;
	sec
	rcall	strb
	brcs	iotimeout
;
; Lies Ergebnis auf OUT0-6
;
	cbi	PORTD, PD4			; '574 Output Enable
	nop
	nop
	in	w2, PINB
	sbi	PORTD, PD4			; '574 Output Disable
	andi	w2, $7f				; /STROBE Bit loeschen
;
; Ende IN-OUT-UNTERPROGRAMM
;
	ret
;
; Timer 1 -> Normal Mode, interner Takt / 1024
; Zähler auf -9000 für 500ms
;
start500ms:
	ldi	w4, high(-9000)
	ldi	w3, low(-9000)
	out	TCNT1H, w4
	out	TCNT1L, w3

	ldi	w4, (1<<TOV1)
	out	TIFR, w4			; Überlauf-Flag löschen

	ldi	w4, (1<<CS12) | (1<<CS10)	; Counter starten
	out	TCCR1B, w4
	ret

check500ms:
	in	w4, TIFR
	andi	w4, (1<<TOV1)
	ret

delay500ms:
	rcall	start500ms
del1:
	rcall	check500ms
	breq	del1
;	rcall	stop500ms
;	ret

stop500ms:
	out	TCCR1B, zero
	ret

putcmd:
	out	PORTB, w1			; Befehl an PortB
	brcc	putcmd1
	sbi	PORTB, PB7			; /READY -> HIGH
putcmd1:
	out	DDRB, ff			; PortB -> Ausgabe
	sbi	PORTD, PD6			; Uebergabe an '574
	cbi	PORTD, PD6
	out	DDRB, zero			; PortB -> Eingabe
	out	PORTB, 	zero			; disable Pull-Up
	ret

strb:
	clr	w2
	brcs	strb1
	ldi	w2, STROBE
strb1:
	rcall	start500ms
strb2:
	cbi	PORTD, PD4			; '574 Output Enable
	nop
	nop
	in	w3, PINB
	sbi	PORTD, PD4			; '574 Output Disable
	andi	w3, STROBE			; nur /STROBE
	eor	w3, w2
	clc
	brne	strb3
	rcall	check500ms
	breq	strb2
	sec
strb3:
;	rcall	stop500ms
;	ret
	rjmp	stop500ms

;
; 8 Bit Division d / n
; -> q = Quotient, r = Rest
;
div8:
	clr	r
	clr	q
	ldi	cnt, 8
div8a:
	lsl	d
	rol	r
	lsl	q
	cp	r, n
	brlo	div8b
	sub	r, n
	inc	q
div8b:
	dec	cnt
	brne	div8a
	ret
;
; 16 Bit Division dh:dl / nh:nl
; -> qh:ql = Quotient, rh:rl = Rest
;
div16:
	clr	rl
	clr	rh
	clr	ql
	clr	qh
	ldi	cnt, 16
div16a:
	lsl	dl
	rol	dh
	rol	rl
	rol	rh
	lsl	ql
	rol	qh
	cp	rh, nh
	brlo	div16c
	brne	div16b
	cp	rl, nl
	brlo	div16c
div16b:
	sub	rl, nl
	sbc	rh, nh
	inc	ql
div16c:
	dec	cnt
	brne	div16a
div16d:
	ret

;
; w1 als zwei Dezimalstellen ausgeben
;
pr2num:
	mov	d, w1
	ldi	w1, 10
	mov	n, w1
	rcall	div8
	mov	w1, q
	rcall	prnum
	mov	w1, r
;	rcall	prnum
;	ret
;
; w1 als ASCII-Ziffer ausgeben
;
prnum:
	andi	w1, $0f
	ori	w1, '0'
;	rcall	tx
;	ret
	rjmp	tx
;
; Zeichenkette an serieller Schnittstelle ausgeben
; Ende mit $00 -> zusätzlich Zeilenende
; Sonderbehandlung; LF wird als CR LF gesendet
;
print1:
	rcall	newline
print:
	lpm	w1, Z+
	tst	w1
	breq	newline
	cpi	w1, LF
	breq	print1
	rcall	tx
	rjmp	print

newline:
	ldi	w1, CR
	rcall	tx
	ldi	w1, LF
;	rjmp	tx

tx:
	sbis	UCSRA, UDRE
	rjmp	tx
	out	UDR, w1
	ret

help1:
	.db	LF, "? -> Menue", 0

help2:
	.db	LF, "Befehl	Betriebsart", LF, "$00	Frequ."
	.db	LF, "$01	Per.", LF, "$02	Zeit", LF, "$03	Tastv.", LF
	.db	"$04	Impuls", LF, LF, "Befehl	Messzeit", LF
	.db	"$10	0,1s", LF, "$11	0,2s", LF, "$12	0,5s", LF, "$13	1s", LF
	.db	"$14	2s", LF, "$15	5s", LF, "$16	10s", LF, "$17	20s", LF
	.db	"$18	50s", LF, "$19	100s", LF, "$1A	200s", LF
	.db	"$1B	500s ", LF, "$1C	1000s", LF, "$1D	2000s", LF
	.db	"$1E	5000s", LF, "$1F	10000s", LF, LF
	.db	"Befehl	Lesen ", LF, "$20	READY-Flag, $F=Not ready, 0=Ready", LF
	.db	"$22	1. Stelle", LF, "$23	2. Stelle", LF, "$24	3. Stelle", LF
	.db	"$25	4. Stelle", LF, "$26	5. Stelle", LF, "$27	6. Stelle", LF
	.db	"$28	7. Stelle", LF, "$29	8. Stelle", LF
	.db	"$2A	4 MSB Exponent", LF, "$2B	4 LSB Exponent", LF
	.db	"$2D	Mantissen-Vorzeichen, 0=Positiv, 1=Negativ", LF, LF
	.db	"Befehl	Aktion", LF, "$30	READY Flag loeschen", LF
	.db	"$3f	Menue", LF, "$40	Dauermessung", LF, 0
